{ Turbo Pascal 6.0
  * HPHP48-? header is expected for all hp object input files
  * This header is also supplied for all hp object output files
  * All output is sent to standard, use > redirecton to save as a file

    HPCON <option> <filename>     output to standard
    -s   hp String object --> pc textfile (^J --> ^M^J)
    -t   pc Textfile      --> hp string object (^M^J --> ^J)
    -g   hp Grob object   --> sasm assembly source
    -G   does -g and inverts grob
    -T   pc Textfile      --> pc textfile
         note: -T fixes textfile for kermit ascii transfer
               of strings with embedded quotes.
               It inserts ''C$ nnn'' between the %%HP header
               and body of text.  Do not quote the text body!

  Randy Ding, Jan 1994
}

{$M 16384,0,65536 }
{$D-}
{$R-}
{$V-}
{$N-}
{$B-}

{$S+}
{$I+}

program hpcon;

const
  header: string[8]= 'HPHP48-E';
  TXTheader: string[5]= '%%HP:';
  strprolog: longint= 10796;
  grobprolog: longint= 11038;
  lonib= false;
  hinib= true;
  maxlongbuf= 64999;

type
  nib= 0..15;
  longbuftype= array[0..maxlongbuf] of byte;

var
  infile: file of byte;
  readbuf,writebuf: byte;          { 1 byte (2 nibble) data buffers }
  readnum,writenum: longint;       { number of nibs read or written }
  readplace,writeplace: boolean;   { either lonib or hinib }
  readempty: boolean;              { true when read buffer empty }
  writefull: boolean;              { true when write buffer full }
  version: char;                   { rom version letter of HP file read }
  longbuf: ^longbuftype;
  longbufpos: word;                { pointer to next char to write }



function fileexists(filename: string): boolean;
  var f: file;
  begin
    {$I-}
    assign (f,filename);
    reset (f);
    close (f);
    {$I+}
    fileexists:= (ioresult=0) and (filename<>'');
  end;



procedure resetoutput;
  begin
    close(output);
    assign(output,'con');
    rewrite(output);
  end;



procedure doerror(s: string);
  begin
    resetoutput;
    writeln;
    write(chr(7));
    writeln(s);
    writeln;
  end;



procedure testfor(filename: string);
  begin
    if not fileexists(filename) then begin
      doerror('file not found');
      halt(1);
    end;
  end;



{ initread must be called before any nibble reads: read1,read2,read5 }
procedure initread;
  begin
    readplace:= lonib;
    readempty:= true;
    readnum:= 0;
    readbuf:= 0;
  end;



{ initwrite must be called before any nibble writes: write1,write2,write5 }
procedure initwrite;
  begin
    writeplace:= lonib;
    writefull:= false;
    writenum:= 0;
    writebuf:= 0;
  end;



{ reads 1 nibble from infile }
{ returns successful flag }
function read1(var x: nib): boolean;
  label exitfunc;
  begin
    if readempty then begin
      if eof(infile) then begin
        read1:= false;
        goto exitfunc;
      end;
      read(infile,readbuf);
      readempty:= false;
      readplace:= lonib;
    end;
    case readplace of
      lonib: begin
               x:= readbuf and 15;
               readplace:= hinib;
             end;
      hinib: begin
               x:= (readbuf shr 4) and 15;
               readempty:= true;
             end;
    end;
    read1:= true;
    readnum:= readnum + 1;
    exitfunc:
  end;



{ does nothing if write buffer empty }
{ returns successful flag }
function flushwritebuf: boolean;
  label exitfunc,skip;
  begin
    case writefull of
      true:  if writeplace = lonib then begin
               flushwritebuf:= false;
               goto exitfunc;
             end;
      false: case writeplace of
               lonib: goto skip;
               hinib: writebuf:= writebuf and 15;
             end;
    end;
    write(chr(writebuf));    { to standard output }
    writefull:= false;
    writeplace:= lonib;
    skip:
    flushwritebuf:= true;
    exitfunc:
  end;



{ write 1 nibble }
{ ! use flushwritebuf before exiting program ! }
{ returns successful flag }
function write1(x: nib): boolean;
  label exitfunc;
  begin
    if writefull then
      if not flushwritebuf then begin
        write1:= false;
        goto exitfunc;
      end;
    case writeplace of
      lonib: begin
               writebuf:= x and 15;
               writeplace:= hinib;
             end;
      hinib: begin
               writebuf:= (writebuf and 15) or ((x shl 4) and 240);
               writefull:= true;
             end;
    end;
    write1:= true;
    writenum:= writenum + 1;
    exitfunc:
  end;



{ read 5 nibs }
{ returns successful flag }
function read5(var x: longint): boolean;
  var
    success: boolean;
    power: longint;
    i: integer;
    j: nib;

  begin
    x:= 0;
    i:= 1;
    power:= 1;
    success:= true;
    while (i<=5) and success do begin
      success:= read1(j);
      x:= x + power * j;
      power:= power * 16;
      i:= i + 1;
    end;
    read5:= success;
  end;



{ write 5 nibs }
{ ! use flushwritebuf before exiting program ! }
{ returns successful flag }
function write5(x: longint): boolean;
  var
    success: boolean;
    i: integer;

  begin
    if (x>1048575) or (x<0) then success:= false
    else begin
      success:= true;
      i:= 1;
      while (i<=5) and success do begin
        success:= write1(x mod 16);
        x:= x div 16;
        i:= i + 1;
      end;
    end;
    write5:= success;
  end;



{ read 2 nibs }
{ returns successful flag }
function read2(var x: byte): boolean;
  var
    success: boolean;
    power: byte;
    i: integer;
    j: nib;

  begin
    x:= 0;
    i:= 1;
    power:= 1;
    success:= true;
    while (i<=2) and success do begin
      success:= read1(j);
      x:= x + power * j;
      power:= lo(power * 16);
      i:= i + 1;
    end;
    read2:= success;
  end;



{ write 2 nibs }
{ ! use flushwritebuf before exiting program ! }
{ returns successful flag }
function write2(x: byte): boolean;
  var
    success: boolean;
    i: integer;

  begin
    success:= true;
    i:= 1;
    while (i<=2) and success do begin
      success:= write1(x mod 16);
      x:= x div 16;
      i:= i + 1;
    end;
    write2:= success;
  end;



{ returns true if header and expected prolog is found }
{ readplace,readempty,readnum must allready be initialized by caller }
function testprolog(expect: longint): boolean;
  var
    success: boolean;
    hstr: string[8];
    i: integer;
    j: byte;
    p: longint;

  begin
    version:= '?';
    success:= true;
    i:= 1;
    hstr:= '';
    while (i<=8) and success do begin
      if eof(infile) then success:= false
      else begin
        read(infile,j);
        hstr:= hstr + chr(j);
        i:= i + 1;
      end;
    end;
    if success then success:= copy(header,1,7) = copy(hstr,1,7);
    if success then version:= hstr[8];
    if success then success:= read5(p);
    if success then success:= p=expect;
    testprolog:= success;
  end;



{ converts hp string object to pc textfile }
{ replaces ^J with ^M^J }
procedure str2text;
  label closeexit;
  var
    success: boolean;
    i,lenfield: longint;
    x: byte;

  begin
    testfor(paramstr(2));
    assign(infile,paramstr(2));
    reset(infile);
    initread;
    if not testprolog(strprolog) then begin
      doerror('error reading header or string prolog');
      goto closeexit;
    end;
    success:= read5(lenfield);
    if (not success) or ((lenfield-5) mod 2 <> 0) then begin
      doerror('error reading string length field');
      goto closeexit;
    end;
    i:= (lenfield-5) div 2;
    while (i>0) and success do begin
      success:= read2(x);
      if success then begin
        if x=10 then write(chr(13));
        write(chr(x));
        i:= i - 1;
      end;
    end;
    if (not success) or (readnum-5 <> lenfield) then begin
      doerror('string length mismatch');
      writeln('actual   : ',(readnum-10)/2);
      writeln('expected : ',(lenfield-5)/2);
      writeln;
    end;
    closeexit:
    close(infile);
  end;



{ converts pc textfile to hp string object }
{ replaces ^M^J with ^J }
procedure text2str;
  label skip,dispexit;
  var
    fs: longint;
    i: integer;
    success,ctrlM: boolean;
    x: byte;

  begin
    testfor(paramstr(2));
    assign(infile,paramstr(2));
    reset(infile);
    new(longbuf);
    longbufpos:= 0;
    ctrlM:= false;
    fs:= filesize(infile);
    success:= true;
    while (fs>0) and success do begin
      success:= longbufpos <= maxlongbuf;
      if success then success:= not eof(infile);
      if success then begin
        read(infile,x);
        case x of
          10: ctrlM:= false;
          13: if not ctrlM then begin
                ctrlM:= true;
                goto skip;
              end;
        else
          if ctrlM then begin
            ctrlM:= false;
            longbuf^[longbufpos]:= 13;
            longbufpos:= longbufpos + 1;
          end;
        end;
        if success then success:= longbufpos <= maxlongbuf;
        if success then begin
          longbuf^[longbufpos]:= x;
          longbufpos:= longbufpos + 1;
        end;
        skip:
      end;
      fs:= fs - 1;
    end;
    if success and ctrlM then begin
      success:= longbufpos <= maxlongbuf;
      if success then begin
        longbuf^[longbufpos]:= x;
        longbufpos:= longbufpos + 1;
        ctrlM:= false;
      end;
    end;
    close(infile);
    if not success then begin
      doerror('write conversion or too large error');
      goto dispexit;
    end;

    initwrite;
    write(header);
    success:= write5(strprolog);
    if success then success:= write5(longbufpos*2+5);
    i:= longbufpos;
    while (i>0) and success do begin
      success:= write2(longbuf^[longbufpos-i]);
      i:= i - 1;
    end;
    if success then success:= flushwritebuf;
    if not success then begin
      doerror('write error');
      goto dispexit;
    end;
    if writenum-10 <> longbufpos*2 then
      doerror('write length mismatch error');
    dispexit:
    dispose(longbuf);
  end;



procedure text2text;
  label procexit,dispexit;
  var
    f: file;
    success,done: boolean;
    headstring: string;
    i,j: word;
    x: byte;

  begin
    testfor(paramstr(2));
    assign(f,paramstr(2));
    reset(f,1);
    done:= eof(f);
    i:= sizeof(headstring);
    success:= true;
    headstring:= '';
    while (not done) and (i>0) and success do begin
      blockread(f,x,1,j);
      if j <> 1 then success:= false;
      if not (chr(x) in [chr(10),chr(13)]) then
        headstring:= headstring + chr(x);
      i:= i - 1;
      done:= x = 10;
    end;
    if (copy(headstring,1,length(TXTheader)) <> TXTheader)
       or (not success) then begin
      doerror('input file header bad or not readable');
      goto procexit;
    end;
    if eof(f) then begin
      doerror('no data following header error');
      goto procexit;
    end;
    new(longbuf);
    blockread(f,longbuf^,sizeof(longbuf^),j);
    if (not eof(f)) or (j=0) then begin
      doerror('input file bad or too large error');
      goto dispexit;
    end;
    writeln(headstring);
    write('C$ ',j,' ');
    if j>0 then for i:= 0 to j-1 do write(chr(longbuf^[i]));
    dispexit:
    dispose(longbuf);
    procexit:
    close(f);
  end;



{ converts hp grob object to hp assembly source file }
procedure grob2asm(invrt: boolean);
  label closeexit;
  var
    i,lenfield,height,width,calclen: longint;
    success: boolean;
    x: nib;
    lineplace: integer;

  begin
    testfor(paramstr(2));
    assign(infile,paramstr(2));
    reset(infile);
    initread;
    if not testprolog(grobprolog) then begin
      doerror('error reading header or grob prolog');
      goto closeexit;
    end;
    success:= read5(lenfield);
    if success then success:= read5(height);
    if success then success:= read5(width);
    if not success then goto closeexit;
    calclen:= width div 4;
    if width mod 4 <> 0 then calclen:= calclen + 1;
    if odd(calclen) then calclen:= calclen + 1;
    calclen:= calclen * height + 15;
    if calclen <> lenfield then begin
      doerror('length field does not match grob dimensions');
      goto closeexit;
    end;
    writeln('ASSEMBLE');
    writeln('    CON(5)  =DOGROB');
    writeln('    CON(5)  ',lenfield);
    writeln('    CON(5)  ',height);
    write  ('    CON(5)  ',width);
    i:= lenfield-15;
    lineplace:= 0;
    while (i>0) and success do begin
      if lineplace=0 then begin
        writeln;
        write('    NIBHEX  ');
      end;
      success:= read1(x);
      if success then begin
        if invrt then x:= x xor 15;
        case (x and 15) of
          0..9   : write(chr(ord('0')+x));
          10..15 : write(chr(ord('A')+x-10));
        end;
        lineplace:= lineplace + 1;
        if lineplace = 64 then lineplace:= 0;
        i:= i - 1;
      end;
    end;
    if success then begin
      writeln;
      writeln('RPL');
    end;
    if not success then
      doerror('grob body read or length error');
    closeexit:
    close(infile);
  end;



procedure showhelp;
  begin
    resetoutput;
    writeln('HPCON <option> <filename>     output to standard');
    writeln;
    writeln('-s   hp String object --> pc textfile (^J --> ^M^J)');
    writeln('-t   pc Textfile      --> hp string object (^M^J --> ^J)');
    writeln('-g   hp Grob object   --> sasm assembly source');
    writeln('-G   does -g and inverts grob');
    writeln('-T   pc Textfile      --> pc textfile');
    writeln('     note: -T fixes textfile for kermit ascii transfer');
    writeln('           of strings with embedded quotes.');
    writeln('           It inserts ''C$ nnn'' between the %%HP header');
    writeln('           and body of text.  Do not quote the text body!');
    writeln('Randy Ding, Jan 1993, FreeWare');
    writeln('Please distribute with source code!');
  end;



procedure main;
  label exitpr;
  var s: string[2];
  begin
    if paramcount < 2 then begin
      showhelp;
      goto exitpr;
    end;
    s:= paramstr(1);
    if s[1] <> '-' then begin
      showhelp;
      goto exitpr;
    end;
    case s[2] of
      's': str2text;
      't': text2str;
      'g': grob2asm(false);
      'G': grob2asm(true);
      'T': text2text;
    else showhelp;
    end;
    exitpr:
  end;



begin
  main;
end.
